#!/usr/bin/env python
 
from setuptools import setup
 
setup(name='brubeck-service',
      version='0.1.2',
      description='Library to allow Brubeck framework to call other instances as services',
      author='Seth Murphy',
      author_email='seth@brooklyncode.com',
      url='http://github.com/sethmurphy/brubeckservices',
      packages=['brubeckservice'])
