#!/usr/bin/env python

"""The Brubeck Services Addon."""

version = "0.1.0"
version_info = (0, 1, 0)
__all__ = ['base, connections, coro, handlers, models, resource, service, ServiceError, tnetstrings']
